package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.ALASKA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.CASCADIA;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.GLOBAL;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.SLAB;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.max;
import static java.lang.Math.pow;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable.Position;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the PEER NGA-Subduction GMM, Kuehn et al., 2020 GMM
 * (KBCG20)
 *
 * Notes:
 *
 * This model is only applicable to sites in the forearc, future residuals
 * analysis is planned to determine if there are attenuation differences between
 * forearc and backarc zones.
 *
 * PSHAB20 uses zHyp (randomized source location on fault), but this can be
 * replaced this event-specific zHyp with the mean depth expected for the fault
 * plane, zHypBar, in Eq. 4.6 without modification to coefficients or
 * uncertainty.
 *
 * Path term (Eqs. 4.2 and 4.3 in PEER report) appears to differ in the official
 * Python implementation. Compare implementations of getPathTerm() below
 *
 * <p><b>Reference:</b> Kuehn, N., Bozorgnia, Y, Campbell, K.W., Gregor, N.
 * (2020) Partially non-ergodic ground-motion model for subduction regions using
 * the NGA-Subduction database. Report no. 2020/04. Berkeley, CA: PEER, 147 p.
 *
 * <p><b>Component:</b> average horizontal (RotD50)
 *
 * Questions:
 *
 * * δMbreak, defined in Figure 4.2. The paper does not mention regionalization
 * for this parameter, but in the Python code only Japan and South America have
 * non-zero values.
 *
 * * Source Zref(if|slab) are not specified in the paper. Python code specifies
 * 15 and 50 km, respectively
 *
 * * Source-Depth breakpoints
 *
 * * Eq. 4.2 - median PSA at short periods is not allowed to be less than PGA,
 * so, if PSA(T&lt;0.1) &lt; PGA, do we still use sigma(PSA(T&lt;0.1)) or use
 * sigma(PGA)?
 *
 * * Eq. 4.8 in the paper includes `c ln( vs30/k1)^n`, this ln is omitted in the
 * Python and the other papers using this term (CB14, etc.)
 *
 * * I think fSite term in Python code has a typo, using `vs30 &lt; k1` rather
 * than `vs30 ≤ k1` as specified in the paper (and others using the same site
 * term)
 *
 * * Value of k2 (hard-coded in Python) for PGV is a typo, the value should be
 * -1.955 but -1.995 is used
 *
 * * Basin Zref is only calculated (coefficients only provided) for CASC, JP,
 * NZ, and TW? Then is Zref == 0 for all other regions?
 *
 * * Basin term for Cascadia Other PacNW basins, the paper does not allow the
 * basin term to exceed that for the Seattle basin. The Python code does not
 * seem to impose this constraint but I may not be seeing it.
 *
 * * Uncertainty values obtained from:
 * https://github.com/nikuehn/KBCG20/tree/master/UNCERTAINTY
 *
 * @author U.S. Geological Survey
 */

public abstract class KuehnEtAl_2020 implements GroundMotionModel {

  static final String NAME = "Kuehn et al. (2020)";

  /*
   * Constraints for magnitude breakpoint, source depth (as noted below),
   * forearc/backarc subregions, basin depth are region-specific
   */
  static final Constraints CONSTRAINTS_INTERFACE = Constraints.builder()
      .set(MW, Range.closed(5.0, 9.5))
      .set(RRUP, Range.closed(10.0, 1000.0))
      .set(ZTOR, Range.closed(0.0, 50.0))
      .set(VS30, Range.closed(150.0, 1500.0))
      .build();
  static final Constraints CONSTRAINTS_SLAB = Constraints.builder()
      .set(MW, Range.closed(5.0, 8.5))
      .set(RRUP, Range.closed(10.0, 1000.0))
      .set(ZTOR, Range.closed(0.0, 200.0)) // for Columbia: ZTOR <= 150
      .set(VS30, Range.closed(150.0, 1500.0))
      .build();

  // static final class Basin {
  // // Other region
  // static final String NONE = "None";
  // // PacNW outside of basins
  // static final String PACNW_NONE = "PacNW-no-basin";
  // // Tacoma/Everett basins
  // static final String OTHER = "Other";
  // // Seattle basin
  // static final String SEATTLE = "Seattle";
  // }

  private static final String KBCG20_CSV = "KBCG20.csv";

  /*
   * Define coefficient containers
   */
  static final CoefficientContainer COEFFS;

  static {
    COEFFS = new CoefficientContainer(KBCG20_CSV);
  }

  /*
   * Define constants
   */
  private static final double δM = 0.1; // p. 20
  private static final double MREF = 6.0; // p. 20
  private static final double δZ = 1.0; // p. 22
  private static final double C = 1.88; // p. 22 !!! Value from python code
  private static final double N = 1.18; // p. 22 !!! Value from python code
  private static final double VSROCK = 1100;

  private static final class Coefficients {
    final Imt imt;

    final double θ1;
    final double θ2;
    final double θ3;
    final double θ4;
    final double θ5;
    final double θ6_2;
    final double θ7;
    final double θ9;
    final double θ10;
    final double θ11;
    final double θ11S;
    final double θ12;

    final double θnft1;
    final double θnft2;

    final double k1;
    final double k2;

    final double φ;
    final double τ;

    final double Mbreak;

    /**
     * <pre>
     * Mb (different from Mbreak???) see getMbDefault() in python code
     *            global Aleutian Alaska Cascadia
     * Interface  7.6     8.0       7.2     7.2
     * Intraslab  7.9     8.0       8.6     8.0
     * </pre>
     */
    final double Zbreak; // dzb_if dzb_slab
    final double Zref;

    Coefficients(Imt imt, CoefficientContainer cc, SubductionZone zone, boolean slab) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);

      // No regionalization or interface/slab
      θ3 = coeffs.get("theta_3");
      θ5 = coeffs.get("theta_5");

      if (slab) {
        θ2 = coeffs.get("theta_2_slab");
        θ4 = coeffs.get("theta_4_slab");
        θ9 = coeffs.get("theta_9_slab");

        Zbreak = 80.0 + coeffs.get("dzb_slab");
        Zref = 50.0; // this value is not in the paper
      } else {
        // set interface coefs
        θ2 = coeffs.get("theta_2_if");
        θ4 = coeffs.get("theta_4_if");
        θ9 = coeffs.get("theta_9_if");

        Zbreak = 30.0 + coeffs.get("dzb_if");
        Zref = 15.0; // this value is not in the paper
      }

      /**
       * ***** mu_theta_6[b] (global) vs. theta_6xc vs. regional theta_6_*_reg,
       * especially for Alaska/Cascadia, which don't seem to use these coeffs
       *
       * <pre>
       *              Global        Alaska              Cascadia
       * theta_6_x1   mu_theta_6b   theta_6_x1_reg_Al   theta_6_x1_reg_Ca
       * theta_6_x2   mu_theta_6b   theta_6_x2_reg_Al   theta_6_x2_reg_Ca
       * theta_6_x3   mu_theta_6b   theta_6_x3_reg_Al   theta_6_x3_reg_Ca
       * theta_6_1    mu_theta_6b   theta_6_1_reg_Al    theta_6_1_reg_Ca
       * theta_6_2    mu_theta_6    theta_6_2_reg_Al    theta_6_2_reg_Ca <-- only one used ???
       * theta_6_3    mu_theta_6b   theta_6_3_reg_Al    theta_6_3_reg_Ca
       * </pre>
       */

      /* For now we apply general Cascadia basin scaling elsewhere. */
      θ11 = coeffs.get("intercept_Ca");
      θ11S = coeffs.get("mean_residual_Seattle_basin");
      θ12 = coeffs.get("slope_Ca");

      if (zone == CASCADIA) {
        θ1 = (slab) ? coeffs.get("theta_1_slab_reg_Ca") : coeffs.get("theta_1_if_reg_Ca");
        θ6_2 = coeffs.get("theta_6_2_reg_Ca");
        θ7 = coeffs.get("theta_7_reg_Ca");

        // if (basin == Basin.SEATTLE) {
        // θ11 = 0;
        // θ11S = coeffs.get("mean_residual_Seattle_basin");
        // θ12 = 0;
        // } else {
        // θ11 = coeffs.get("intercept_Ca");
        // θ11S = coeffs.get("mean_residual_Seattle_basin");
        // θ12 = coeffs.get("slope_Ca");
        // }

      } else if (zone == ALASKA) {

        θ1 = (slab) ? coeffs.get("theta_1_slab_reg_Al") : coeffs.get("theta_1_if_reg_Al");
        θ6_2 = coeffs.get("theta_6_2_reg_Al");
        θ7 = coeffs.get("theta_7_reg_Al");
        // θ11 = 0;
        // θ11S = 0;
        // θ12 = 0;

      } else {
        // global
        θ1 = (slab) ? coeffs.get("mu_theta_1_slab") : coeffs.get("mu_theta_1_if");
        θ6_2 = coeffs.get("mu_theta_6");
        θ7 = coeffs.get("mu_theta_7");
        // θ11 = 0;
        // θ11S = 0;
        // θ12 = 0;
      }

      θ10 = 0.0; // Table 4.1: fixed to 0
      θnft1 = coeffs.get("nft_1");
      θnft2 = coeffs.get("nft_2");

      k1 = coeffs.get("k1");
      k2 = coeffs.get("k2");

      Mbreak = getBreakpointMag(imt, zone, slab);
      // δZb = (isSlab) ? coeffs.get("dzb_slab") : coeffs.get("dzb_if");

      φ = coeffs.get("phi");
      τ = coeffs.get("tau");
    }

    /*
     * Breakpoint magnitude, Mb, and δMbreak for the magnitude term (Eq. 4.4).
     *
     * The paper does not mention regionalization for δMbreak, but the Python
     * code only applies δMbreak for Japan and South America
     */
    private double getBreakpointMag(Imt imt, SubductionZone zone, boolean slab) {
      // breakpoint magnitudes are hard coded in python code, see getMbDefault()

      // global
      double Mb = slab ? 7.6 : 7.9;
      if (zone == CASCADIA)
        Mb = slab ? 7.2 : 8.0;
      else if (zone == ALASKA) {
        Mb = slab ? 7.2 : 8.6;
      }

      return Mb;

      /*
       * Figure 4.2: (only for Interface) δMbreak = 0.0 for T<=1.0s, -0.4 for
       * T>=3.0s, and log-linear interpolated in between
       *
       * ??? Python code only calculates this for Japan and South America,
       * δMbreak is 0.0 for all other regions. The paper makes no mention of
       * this term being regionalized
       */
      // double δMbreak = 0.0;
      // if (!slab) {
      // if (!imt.isSA() || imt.period() <= 1.0) {
      // δMbreak = 0.0;
      // } else if (imt.period() >= 3.0) {
      // δMbreak = -0.4;
      // } else {
      // δMbreak = 0.0 + (log(imt.period()) - log(1.0)) * (-0.4 - 0.0) /
      // (log(3.0) - log(1.0));
      // }
      // }
      // // To match Python code:
      // switch (region) {
      // case Region.JAPAN:
      // case Region.SA:
      // return Mb + δMbreak;
      // default:
      // return Mb;
      // }
    }
  }

  private final Coefficients coeffs;
  private final Coefficients coeffsPGA;
  private final GroundMotionTable epistemicTable;

  KuehnEtAl_2020(Imt imt, SubductionZone zone) {
    coeffs = new Coefficients(imt, COEFFS, zone, slab());
    coeffsPGA = new Coefficients(PGA, COEFFS, zone, slab());
    epistemicTable = GroundMotionTables.getKbcg20(
        zone,
        (slab()) ? SubductionZone.Type.SLAB : SubductionZone.Type.INTERFACE,
        imt);
  }

  abstract boolean slab();

  abstract boolean basin();

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {

    // Calc PGAref
    double pgaRef = calcPgaRef(coeffsPGA, in.Mw, in.rRup, in.zTor); // exp(calcMean());
    pgaRef = exp(pgaRef);

    double μ = calcMean(coeffs, basin(), in.Mw, in.rRup, in.zTor, in.vs30, in.z2p5, pgaRef);

    /*
     * Eq. 4.2 - median PSA at short periods is not allowed to be smaller than
     * PGA
     *
     * TODO: If PSA(T<0.1) < PGA, do we still use sigma(PSA(T<0.1)) or use
     * sigma(PGA)?
     */
    if (coeffs.imt.isSA() && coeffs.imt.period() <= 0.1) {
      // calculate PGA
      double μPga =
          calcMean(coeffsPGA, basin(), in.Mw, in.rRup, in.zTor, in.vs30, in.z2p5, pgaRef);
      // μ can't be smaller than PGA
      μ = max(μ, μPga);
      // sigma is not period-dependent, so no change to sigma
    }

    double sigmaAleatory = Maths.hypot(coeffs.φ, coeffs.τ);

    // epistemic uncertainty
    // double sigmaEpistemic = coeffs.ψ; // compute by sampling posterior
    // results

    // epistemic uncertainty from pre-computed tables
    Position p = epistemicTable.position(in.rRup, in.Mw);
    double ψ = epistemicTable.get(p);

    // double sigmaTotal = Maths.hypot(coeffs.φ, coeffs.τ, sigmaEpistemic);

    double σ = sigmaAleatory;

    // return GroundMotions.createTree(μ, σ);

    double[] μs = new double[] { μ + ψ, μ, μ - ψ };

    // return GroundMotions.createTree(μ, σ);
    return GroundMotions.createTree(GroundMotions.EPI_IDS, μs, GroundMotions.EPI_WTS, σ);
  }

  private static double calcPgaRef(
      Coefficients c,
      double Mw,
      double rRup,
      double zTor) {
    double fMag = getMagTerm(c, Mw);

    double fGeom = getGeomAttenTerm(c, Mw, rRup);

    double fDepth = getDepthTerm(c, zTor);

    double fAtten = getArcCrossingTerm(c, rRup);

    double fSite = getSiteAmpTerm(c, VSROCK, 0);

    double fBasin = 0;

    // System.out.printf("pgaRef fMag = %.8f\n", fMag);
    // System.out.printf("pgaRef fGeom = %.8f\n", fGeom);
    // System.out.printf("pgaRef fDepth = %.8f\n", fDepth);
    // System.out.printf("pgaRef fAtten = %.8f\n", fAtten);
    // System.out.printf("pgaRef fSite = %.8f\n", fSite);
    // System.out.printf("pgaRef fBasin = %.8f\n", fBasin);

    return c.θ1 + fMag + fGeom + fDepth + fAtten + fSite + fBasin;
  }

  private static double calcMean(
      Coefficients c,
      boolean basin,
      double Mw,
      double rRup,
      double zTor,
      double vs30,
      double z2p5,
      double pgaRef) {

    double fMag = getMagTerm(c, Mw);

    double fGeom = getGeomAttenTerm(c, Mw, rRup);

    double fDepth = getDepthTerm(c, zTor);

    double fAtten = getArcCrossingTerm(c, rRup);

    double fSite = getSiteAmpTerm(c, vs30, pgaRef);

    double fBasin = basin ? getBasinTerm(c, vs30, z2p5) : 0.0;

    // System.out.printf("calcMean fMag = %.8f\n", fMag);
    // System.out.printf("calcMean fGeom = %.8f\n", fGeom);
    // System.out.printf("calcMean fDepth = %.8f\n", fDepth);
    // System.out.printf("calcMean fAtten = %.8f\n", fAtten);
    // System.out.printf("calcMean fSite = %.8f\n", fSite);
    // System.out.printf("calcMean fBasin = %.8f\n", fBasin);

    return c.θ1 + fMag + fGeom + fDepth + fAtten + fSite + fBasin;
  }

  /*
   * Eq. 4.3 used for mangitude and source-depth terms
   */
  private static double loghinge(double x, double x0, double a, double b0, double b1,
      double δ) {
    return a + b0 * (x - x0) + (b1 - b0) * δ * log(1 + exp((x - x0) / δ));
  }

  /*
   * Eq. 4.4 - magnitude term
   */
  private static double getMagTerm(
      Coefficients c,
      double Mw) {
    return loghinge(Mw, c.Mbreak, c.θ4 * (c.Mbreak - MREF), c.θ4, c.θ5, δM);
  }

  /*
   * Eq. 4.5 - geometrical attenuation term Eq. 4.6 - finite fault term
   * ("pow(10,...)")
   */
  private static double getGeomAttenTerm(
      Coefficients c,
      double Mw,
      double rRup) {
    return (c.θ2 + c.θ3 * Mw) * log(rRup + pow(10, c.θnft1 + c.θnft2 * (Mw - 6.0)));
  }

  /**
   * Eq. 4.7 - source depth term
   *
   * <pre>
   * fDepth(interface|slab) = loghinge(zTor, Zbreak + δZbreak, θ9 * (Zbreak + δZbreak - Zref), θ9, θ10, δZ)
   * </pre>
   *
   * "The source-depth breakpoints [Zbreak + δZbreak] were initially set to
   * Zbreak,if = 30 km and Zbreak,slab = 80 km. The period-dependent adjustment
   * coefficients δZbreak were determined by regression because the depth
   * breakpoints have no theoretical basis... Therefore, the effective source
   * depth breakpoints become Zbreak + δZbreak..."
   *
   * "The parameter δZ determines the smoothness of transition in the bilinear
   * form and was set to δZ = 1."
   *
   * "The coefficients for the source-depth scaling rates (DSR) up to the depth
   * breakpoints are θ9,if and θ9,slab. Both of the coefficients for the DSRs
   * above the breakpoints were fixed at θ10 = 0."
   *
   *
   */
  private static double getDepthTerm(
      Coefficients c,
      double zTor) {
    return loghinge(zTor, c.Zbreak, c.θ9 * (c.Zbreak - c.Zref), c.θ9, c.θ10, δZ);
  }

  /*
   * Eq. 4.8 - site amplification term
   */
  private static double getSiteAmpTerm(Coefficients c, double vs30, double pgaRef) {
    double siteTerm;
    if (vs30 <= c.k1) {
      siteTerm = c.θ7 * log(vs30 / c.k1) +
          c.k2 * (log(pgaRef + C * pow(vs30 / c.k1, N)) - log(pgaRef + C));
    } else {
      siteTerm = (c.θ7 + c.k2 * N) * log(vs30 / c.k1);
    }
    return siteTerm;
  }

  /*
   * Eq. 4.9 - anelastic attenuation term (arc-crossing term) -- Region specific
   * (regionalization ignored for now)
   *
   * Alaska and Cascadia do not have different anelastic attenuation in forearc
   * and backarc
   *
   * Global does have some backarc travel path effects, but these are not yet
   * implemented and we will need to track Ri, the travel distance through
   * subregion i (of the arc) for a given geographic region.
   */
  private static double getArcCrossingTerm(Coefficients c, double rRup) {
    // This is accurate for ALASKA, CASCADIA, NZ, and TW
    return c.θ6_2 * rRup;
  }

  /*
   * Eq. 4.10 - basin depth term
   *
   * Eq. 4.11 - lnZref(vs30)
   *
   * Uses EITHER Z1.0 or Z2.5: Cascadia --> Z2.5, Alaska???
   *
   * TODO: This function follows the paper, it's not clear if it is consistent
   * with the official python.
   */
  private static double getBasinTerm(
      Coefficients c,
      double vs30,
      double z2p5) {

    if (Double.isNaN(z2p5)) {
      return 0.0;
    }

    double z2p5m = z2p5 * 1000.0;
    double δlnZ = log(z2p5m) - getBasinZref(vs30); // getBasinZref(region,
                                                   // vs30);
    /*
     * The Seattle basin, as defined, is relatively limited in range and
     * encloses sites that are all very deep, hence the fixed Seattle basin
     * scale factor. Here, we opt for the general PNW basin (something seems
     * wrong with using the seattle coefficient as a minimum value). (Eqn. 4.14)
     */
    double fb = c.θ11 + c.θ12 * δlnZ;

    return fb; // Math.min(fb, c.θ11S); disabled seattle min cap

    // switch (region) {
    // case Region.CASCADIA:
    // // Eq. 4.14
    // switch (basin) {
    // case Basin.SEATTLE:
    // return c.θ11S; // Seattle-specific
    // case Basin.OTHER:
    // // TODO:
    // // The paper returns this:
    // // System.out.printf(" OTHER basin: taking minimum of %.9f and
    // // %.9f\n",c.θ11 + c.θ12 * δlnZ, c.θ11S);
    // // return min(c.θ11 + c.θ12 * δlnZ, c.θ11S);
    // // The official python code just returns this:
    // return c.θ11 + c.θ12 * δlnZ;
    // default:
    // return 0.0;
    // }
    // default:
    // // Eq. 4.10
    // return c.θ11 + c.θ12 * δlnZ;
    // }
  }

  /*
   * Eq. 4.11
   */
  // private static double getBasinZref(String region, double vs30) {
  private static double getBasinZref(double vs30) {
    /**
     * Eq. 4.11 - lnZref has regional coefs θz1 - θz4 for CASC, JP, NZ, and TW.
     * But then the text says "For CASC, estimating all parameters of Equation
     * 4.11 is not possible due to the very limited number of stations with
     * associated Vs30 and Z2.5 values at hard rock sites in particular. Hence,
     * in this case the model is constrained to approach a value of Zref = 10 m
     * for large values of Vs30 (Vs30 > 2000 m/s)."
     *
     * - Max vs30 for the model is supposed to be 1500 m/s
     *
     * - What coefs θz1 - θz4 for all other regions?
     *
     * - Is Basin-Depth term only applied to CASC, NZ, TW, JP regions?
     *
     * Hard-coded coeffs in Python code, explicitly for calculating delta_ln_z
     * for Cascadia (all basins), for non-PGA-ref calcs, and the first and last
     * values aren't used:
     *
     * <pre>
     * pars_z_casc = [-999, 8.29404964010203, 2.30258509299405, 6.39692965521615, 0.27081459, 1.7381352625]
     * </pre>
     */
    // if (region == Region.CASCADIA) {
    double θz1 = 8.29404964010203;
    double θz2 = 2.30258509299405;
    double θz3 = 6.39692965521615;
    double θz4 = 0.27081459;
    double vsterm = exp((log(vs30) - θz3) / θz4);
    return θz1 + (θz2 - θz1) * vsterm / (1.0 + vsterm);
    // } else {
    // // Python code also has hard-coded coeffs for JA, NZ, and TW
    // return 0.0;
    // }
  }

  /*
   * Calculate Aleatory Sigma
   */
  private static double getAleatorySigma(Coefficients c) {
    return Maths.hypot(c.φ, c.τ);
  }

  private static abstract class Global extends KuehnEtAl_2020 {
    static final SubductionZone ZONE = GLOBAL;
    static final String NAME = KuehnEtAl_2020.NAME + " : " + ZONE;

    Global(Imt imt) {
      super(imt, ZONE);
    }

    @Override
    boolean slab() {
      return false;
    }

    @Override
    boolean basin() {
      return false;
    }
  }

  private static abstract class Cascadia extends KuehnEtAl_2020 {
    static final SubductionZone ZONE = CASCADIA;
    static final String NAME = KuehnEtAl_2020.NAME + " : " + ZONE;

    Cascadia(Imt imt) {
      super(imt, ZONE);
    }

    @Override
    boolean slab() {
      return false;
    }

    @Override
    boolean basin() {
      return false;
    }
  }

  private static abstract class Alaska extends KuehnEtAl_2020 {
    static final SubductionZone ZONE = ALASKA;
    static final String NAME = KuehnEtAl_2020.NAME + " : " + ZONE;

    Alaska(Imt imt) {
      super(imt, ZONE);
    }

    @Override
    boolean slab() {
      return false;
    }

    @Override
    boolean basin() {
      return false;
    }
  }

  static class GlobalInterface extends Global {
    static final String NAME = Global.NAME + " : " + INTERFACE;

    GlobalInterface(Imt imt) {
      super(imt);
    }
  }

  static class GlobalSlab extends Global {
    static final String NAME = Global.NAME + " : " + SLAB;

    GlobalSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static class CascadiaInterface extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + INTERFACE;

    CascadiaInterface(Imt imt) {
      super(imt);
    }
  }

  static class CascadiaSlab extends Cascadia {
    static final String NAME = Cascadia.NAME + " : " + SLAB;

    CascadiaSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

  static final class CascadiaInterfaceBasin extends CascadiaInterface {
    static final String NAME = CascadiaInterface.NAME + " (basin)";

    CascadiaInterfaceBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class CascadiaSlabBasin extends CascadiaSlab {
    static final String NAME = CascadiaSlab.NAME + " (basin)";

    CascadiaSlabBasin(Imt imt) {
      super(imt);
    }

    @Override
    boolean basin() {
      return true;
    }
  }

  static final class AlaskaInterface extends Alaska {
    static final String NAME = Alaska.NAME + " : " + INTERFACE;

    AlaskaInterface(Imt imt) {
      super(imt);
    }
  }

  static final class AlaskaSlab extends Alaska {
    static final String NAME = Alaska.NAME + " : " + SLAB;

    AlaskaSlab(Imt imt) {
      super(imt);
    }

    @Override
    boolean slab() {
      return true;
    }
  }

}
