/**
 *
 */
package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmUtils.BASE_10_TO_E;
import static java.lang.Math.exp;
import static java.lang.Math.log10;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.annotations.Beta;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.gmm.NgaEastUsgs_2017.SiteAmp;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the Pezeshk et al. (2018) ground motion model for central
 * and eastern North America (CENA). Pezeshk et al. (2018) provide coefficients
 * for two hybrid empirical method (HEM) ground-motion prediction equations, one
 * based on a stochastic-scaling approach (PZCT-1SS) and one using the
 * empirical-scaling approach (PZCT-2ES). The latter model is preferred by the
 * authors as the empirical-scaling approach reduces oversaturation of GMIM
 * predictions at large magnitudes, short distances, and short periods.
 *
 * The prediction of GMIMs for site conditions other than Vs30 = 3000 m/s and k0
 * = 0.006s requires the use of appropriate site-amplification factors.
 *
 * Note: Figure 5 in the paper is incorrect. Pezeshk sent a corrected figure
 * that is used to validate the PZCT18-2ES model.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Reference:</b> Pezeshk, S., Zandieh, A., Campbell, K.W., and Tavakoli,
 * B., 2018, Ground-motion prediction equations for Central and Eastern North
 * America using the hybrid empirical method and NGA-West2 Empirical
 * ground-motion models: Bulletin of the Seismological Society of America, v.
 * 108, n. 4, p. 2278-2304, August 2018.
 *
 *
 * <p><b>doi:</b> <a href="http://dx.doi.org/10.1785/0120170179">
 * 10.1785/0120170179</a>
 *
 * <p><b>Component:</b> **** RotD50 ?
 *
 * @author U.S. Geological Survey
 * @see Gmm#NGA_EAST_SEED_PZCT18_1SS
 * @see Gmm#NGA_EAST_SEED_PZCT18_2ES
 */
@Beta
public abstract class PezeshkEtAl_2018 implements GroundMotionModel {

  static final String NAME = "Pezeshk et al. (2018)";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(4.0, 8.0))
      .set(RRUP, Range.closed(0.0, 1000.0))
      .set(VS30, Range.closed(0.0, 3000.0)) // ?
      .build();

  static final class Model {
    static final String PZCT18_1SS = "PZCT18_1SS";
    static final String PZCT18_2ES = "PZCT18_2ES";
  }

  /*
   * Note: 0.04 and 0.08 second coefficents are commented out because NGA-East
   * SiteAmp doesn't support SA0P04 (also commented out there) or SA0P08.
   */
  static final CoefficientContainer COEFFS_1SS = new CoefficientContainer("PZCT18_1SS.csv");
  static final CoefficientContainer COEFFS_2ES = new CoefficientContainer("PZCT18_2ES.csv");
  static final CoefficientContainer COEFFS_SIGMAS = new CoefficientContainer("PZCT18_sigmas.csv");

  private static final class Coefficients {
    final Imt imt;
    final double c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, σReg;
    final double c12, c13, c14, c15, c16, c17, c18, c19, c20, c21, c22, c23, c24, c25;
    final String model;

    Coefficients(Imt imt, CoefficientContainer cc, CoefficientContainer cc_sigma, String model) {
      this.imt = imt;
      this.model = model;

      Map<String, Double> coeffs = cc.get(imt);
      Map<String, Double> coeffs_sigma = cc_sigma.get(imt);

      /*
       * Pezeshk confirmed that the "average empirical calibration factor" of
       * 0.32 (log10-space), discussed on p. 2287, is included in coefficient c1
       * in Table 5 (PZCT18_2ES) and sent a corrected Figure 4 for validation.
       */

      // median model coefficients
      c1 = coeffs.get("c1");
      c2 = coeffs.get("c2");
      c3 = coeffs.get("c3");
      c4 = coeffs.get("c4");
      c5 = coeffs.get("c5");
      c6 = coeffs.get("c6");
      c7 = coeffs.get("c7");
      c8 = coeffs.get("c8");
      c9 = coeffs.get("c9");
      c10 = coeffs.get("c10");
      c11 = coeffs.get("c11");
      σReg = coeffs.get("σReg");

      // between-event variability coefficients
      c12 = coeffs_sigma.get("c12");
      c13 = coeffs_sigma.get("c13");
      c14 = coeffs_sigma.get("c14");
      c15 = coeffs_sigma.get("c15");
      c16 = coeffs_sigma.get("c16");
      c17 = coeffs_sigma.get("c17");
      c18 = coeffs_sigma.get("c18");
      // within-event variability coefficients
      c19 = coeffs_sigma.get("c19");
      c20 = coeffs_sigma.get("c20");
      c21 = coeffs_sigma.get("c21");
      c22 = coeffs_sigma.get("c22");
      c23 = coeffs_sigma.get("c23");
      c24 = coeffs_sigma.get("c24");
      c25 = coeffs_sigma.get("c25");
    }

  }

  private final Coefficients coeffs;
  private final Coefficients coeffsPga;
  private final SiteAmp siteAmp;

  PezeshkEtAl_2018(Imt imt, String model) {
    CoefficientContainer cc = (model == Model.PZCT18_1SS) ? COEFFS_1SS : COEFFS_2ES;
    coeffs = new Coefficients(imt, cc, COEFFS_SIGMAS, model);
    coeffsPga = new Coefficients(Imt.PGA, cc, COEFFS_SIGMAS, model);
    this.siteAmp = new NgaEastUsgs_2017.SiteAmp(imt);
  }

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {
    double μPga = exp(calcMean(coeffsPga, in.Mw, in.rRup));
    double μ = calcMean(coeffs, in.Mw, in.rRup);
    double σ = calcStdDev(coeffs, in.Mw);

    SiteAmp.Value fSite = siteAmp.calc(μPga, in.vs30);
    μ = fSite.apply(μ);

    return GroundMotions.createTree(μ, σ);
  }

  private double calcMean(Coefficients c, double Mw, double Rrup) {
    /* Eq. 5 */
    double R = sqrt(Rrup * Rrup + c.c11 * c.c11);

    /* Eq. 4 - log10 units */
    double μ = c.c1 +
        (c.c2 * Mw) +
        (c.c3 * Mw * Mw) +
        (c.c4 + c.c5 * Mw) * min(log10(R), log10(60.0)) +
        (c.c6 + c.c7 * Mw) * max(min(log10(R / 60.0), log10(2.0)), 0.0) +
        (c.c8 + c.c9 * Mw) * max(log10(R / 120.0), 0.0) +
        (c.c10 * R);
    return μ * BASE_10_TO_E;
  }

  private double calcStdDev(Coefficients c, double Mw) {
    /* Aleatory standard deviation */
    /* Eq. 6 (between-event term) and Eq. 7 (within-event term) - natural log */
    double τ, φ;
    if (Mw <= 4.5) {
      τ = c.c12;
      φ = c.c19 + c.c20 * Mw;
    } else if (Mw <= 5.0) {
      τ = c.c13 + c.c14 * Mw;
      φ = c.c21 + c.c22 * Mw;
    } else if (Mw <= 6.5) {
      τ = c.c15 + c.c16 * Mw;
      φ = c.c23 + c.c24 * Mw;
    } else {
      τ = c.c17 + c.c18 * Mw;
      φ = c.c25;
    }
    return sqrt(τ * τ + φ * φ);
  }

  static final class Seed_PZCT18_1SS extends PezeshkEtAl_2018 {
    static final String MODEL = Model.PZCT18_1SS;
    static final String NAME = PezeshkEtAl_2018.NAME + " : " + MODEL;

    Seed_PZCT18_1SS(Imt imt) {
      super(imt, MODEL);
    }
  }

  static final class Seed_PZCT18_2ES extends PezeshkEtAl_2018 {
    static final String MODEL = Model.PZCT18_2ES;
    static final String NAME = PezeshkEtAl_2018.NAME + " : " + MODEL;

    Seed_PZCT18_2ES(Imt imt) {
      super(imt, MODEL);
    }
  }

}
