package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.CaseFormat.LOWER_HYPHEN;
import static com.google.common.base.CaseFormat.UPPER_UNDERSCORE;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.DipSlipModel.FIXED;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.TREE_INFO;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.checkGridDataDirectory;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.checkSourceTree;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readCalcConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readDecollementConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultClusterSet;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultFeatures;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultRateTree;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readFaultRuptureSet;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readGmms;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readGridConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readGridFeatures;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readGridRuptureSets;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readInterfaceConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readInterfaceFeatures;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readInterfaceRuptureSet;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readMfdConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readMfdMap;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readModelInfo;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readRateTree;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readSiteData;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readSlabConfig;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readSlabRuptureSets;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readSourceTree;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readSystemRuptureSet;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readTreeInfo;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readZoneConfig;
import static gov.usgs.earthquake.nshmp.model.SourceType.DECOLLEMENT;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_SYSTEM;
import static gov.usgs.earthquake.nshmp.model.SourceType.GRID;
import static gov.usgs.earthquake.nshmp.model.SourceType.INTERFACE;
import static gov.usgs.earthquake.nshmp.model.SourceType.SLAB;
import static gov.usgs.earthquake.nshmp.model.SourceType.ZONE;
import static java.nio.file.Files.newDirectoryStream;

import java.io.IOException;
import java.math.RoundingMode;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.function.Supplier;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.google.common.collect.Iterables;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.calc.CalcConfig;
import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.model.HazardModel.MapRegion;
import gov.usgs.earthquake.nshmp.model.SourceFeature.Key;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicGroup;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Earthquake source model loading class.
 *
 * Although source types are broadly similar, differences in model structure
 * necessitate source type specific loaders. Models are loaded recursively and
 * the work of checking the existance of, reading, and deserializing are handled
 * by the factory classes ModelFiles and Deserialize.
 *
 * @author U.S. Geological Survey
 */
abstract class ModelLoader {

  /*
   * Developer notes:
   *
   * In process*Branch methods we are exiting for null ("do nothing") branches
   * (so not adding them to source tree) and therefore can't sample.
   *
   * Need to be clear in documentation about files that are expected at source
   * roots (mfd-map, grid-data/, features) and those that can be specified
   * anywhere in source heirarchy (*-config.json, rate-tree-json). The latter
   * are currently not necessarily being checked at each level of a source tree
   * (e.g. slab and grid check for source-config at processBranch level, but not
   * at sourceTree level.
   */

  public static void main(String[] args) {

    // Path testModel = Paths.get("../nshm-conus-2018-tmp");
    Path testModel = Paths.get("../nshm-conus");
    // Path testModel = Paths.get("../nshm-hawaii");
    HazardModel model = ModelLoader.load(testModel);
    System.out.println();
    System.out.println(model);

  }

  private static Logger log;

  static {
    log = Logger.getLogger(ModelLoader.class.getName());
  }

  static HazardModel load(Path dir) {

    try {
      System.out.println(dir + Text.NEWLINE);

      CalcConfig config = readCalcConfig(dir);

      SiteData siteData = readSiteData(dir);
      System.out.println(siteData);

      HazardModel.Builder builder = HazardModel.builder()
          .info(readModelInfo(dir).orElseThrow())
          .config(config)
          .siteData(siteData);

      /* Process tectonic setting source directories. */
      loadTectonicSettings(dir, builder);

      return builder.build();

    } catch (Exception e) {
      String pathString = (dir != null)
          ? dir.toAbsolutePath().normalize().toString()
          : "null";
      log.log(Level.SEVERE, "Error loading model: " + pathString, e);
      throw e;
    }
  }

  /* Process teconic setting directories. */
  private static void loadTectonicSettings(
      Path dir,
      HazardModel.Builder builder) {

    for (TectonicSetting setting : TectonicSetting.values()) {
      String tsDir = UPPER_UNDERSCORE.to(LOWER_HYPHEN, setting.name());
      Path settingDir = dir.resolve(tsDir);
      if (Files.exists(settingDir)) {
        List<SourceTree> trees = loadTectonicSetting(setting, settingDir, builder);
        trees.forEach(builder::addSourceTree);
      }
    }
  }

  private static List<SourceTree> loadTectonicSetting(
      TectonicSetting setting,
      Path dir,
      HazardModel.Builder builder) {

    Optional<MapRegion> region = ModelFiles.readMapRegion(dir);
    region.ifPresent(mr -> builder.addMapRegion(setting, mr));

    switch (setting) {
      case ACTIVE_CRUST:
        return crustalSources(dir, builder.config, setting);
      case STABLE_CRUST:
        return crustalSources(dir, builder.config, setting);
      case SUBDUCTION:
        return subductionSources(dir, builder.config, setting);
      case VOLCANIC:
        return volcanicSources(dir, builder.config, setting);
      default:
        return List.of();
    }
  }

  private static List<SourceTree> crustalSources(
      Path dir,
      CalcConfig config,
      TectonicSetting setting) {

    List<SourceTree> trees = new ArrayList<>();
    Optional<GmmSet> gmms = ModelFiles.readGmms(dir);
    for (SourceType type : SourceType.values()) {
      Path typeDir = dir.resolve(type.name().toLowerCase());
      if (!Files.exists(typeDir)) continue;
      switch (type) {
        case FAULT:
          trees.addAll(faultSources(typeDir, setting, gmms));
          break;
        case GRID:
          trees.addAll(gridSources(typeDir, setting, gmms));
          break;
        case ZONE:
          trees.addAll(zoneSources(typeDir, setting, gmms));
          break;
        default:
      }
    }
    return trees;
  }

  private static List<SourceTree> subductionSources(
      Path dir,
      CalcConfig config,
      TectonicSetting setting) {

    List<SourceTree> trees = new ArrayList<>();
    Optional<GmmSet> gmms = ModelFiles.readGmms(dir);
    for (SourceType type : SourceType.values()) {
      Path typeDir = dir.resolve(type.name().toLowerCase());
      if (!Files.exists(typeDir)) continue;
      switch (type) {
        case INTERFACE:
          trees.addAll(interfaceSources(typeDir, setting, gmms));
          break;
        case SLAB:
          trees.addAll(slabSources(typeDir, setting, gmms));
          break;
        default:
      }
    }
    return trees;
  }

  private static List<SourceTree> volcanicSources(
      Path dir,
      CalcConfig config,
      TectonicSetting setting) {

    List<SourceTree> trees = new ArrayList<>();
    Optional<GmmSet> gmms = ModelFiles.readGmms(dir);
    for (SourceType type : SourceType.values()) {
      Path typeDir = dir.resolve(type.name().toLowerCase());
      if (!Files.exists(typeDir)) continue;
      switch (type) {
        case GRID:
          trees.addAll(gridSources(typeDir, setting, gmms));
          break;
        case DECOLLEMENT:
          trees.addAll(decollementSources(typeDir, setting, gmms));
          break;
        case ZONE:
          trees.addAll(zoneSources(typeDir, setting, gmms));
          break;
        default:
      }
    }
    return trees;
  }

  /* General purpose NSE supplier for required optionals. */
  private static Supplier<NoSuchElementException> nse(Path dir, String filename) {
    String missingFile = dir.resolve(filename).toString();
    return () -> new NoSuchElementException(missingFile);
  }

  /* --------------- ↑ static - abstract ↓ --------------- */

  final Path root;

  private ModelLoader(Path root) {
    this.root = root;
  }

  /* Load source type specific config, if present. */
  abstract void loadSourceConfig(Path dir, ModelData data);

  /* Load source type specific rate tree, if present. */
  abstract void loadRateTree(Path dir, ModelData data);

  /* Load single (branch) type specific source. */
  abstract SourceTree loadSingleSource(
      Path geojson,
      ModelData data);

  /* Load source type specific tree. */
  abstract SourceTree loadSourceTree(
      Path dir,
      ModelData data);

  /* Source type independent recursive directory processor. */
  List<SourceTree> loadSourceDirectory(
      Path dir,
      ModelData dataIn) {

    /* Update possible overrides in copy of support class. */
    ModelData data = ModelData.copyOf(dataIn);
    readGmms(dir).ifPresent(data::gmms);
    readMfdConfig(dir).ifPresent(data::mfdConfig);
    readMfdMap(dir).ifPresent(data::mfdMap);
    loadSourceConfig(dir, data);
    loadRateTree(dir, data);

    /* (1) Process source tree, if present. */
    if (checkSourceTree(dir).isPresent()) {
      return List.of(loadSourceTree(dir, data));
    }

    List<SourceTree> trees = new ArrayList<>();

    /* (2) Process any GeoJSON source files. */
    ModelFiles.readGeoJsons(dir, data, this::loadSingleSource, trees::add);

    /* (3) Lastly, recurse into any nested directories. */
    ModelFiles.readNestedDirectories(dir, data, this::loadSourceDirectory, trees::addAll);

    return trees;
  }

  private static void logRoot(Path root, SourceType type) {
    int end = root.getNameCount();
    int start = (type == SLAB || type == INTERFACE) ? end - 1 : end - 2;
    System.out.println("\n" + root.subpath(start, end));
  }

  /* ------------ ↑ abstract - implementations ↓ ------------ */

  static List<SourceTree> faultSources(
      Path dir,
      TectonicSetting setting,
      Optional<GmmSet> gmms) {

    GmmSet sourceGmms = readGmms(dir).or(() -> gmms).orElseThrow();
    SourceConfig srcConfig = readFaultConfig(dir).orElseThrow();
    ModelData data = ModelData.create(setting, sourceGmms, srcConfig);
    return new Fault(dir).loadSourceDirectory(dir, data);
  }

  private static class Fault extends ModelLoader {

    Fault(Path root) {
      super(root);
      logRoot(root, FAULT);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readFaultConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readFaultRateTree(dir).ifPresent(data::faultRateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      /* Ensure singleton mfd-map. */
      checkState(data.mfdMap().orElseThrow().size() == 1);

      SourceFeature.NshmFault feature = SourceFeature.newNshmFault(geojson);
      System.out.println("    tree: [" + feature.id + "] " + root.relativize(geojson));

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(feature.name)
          .id(feature.id)
          .setting(data.tectonicSetting())
          .type(FAULT)
          .gmms(data.gmms());

      return loadSingleSource(treeBuilder, Optional.empty(), feature, data).build();
    }

    private SourceTree.Builder loadSingleSource(
        SourceTree.Builder treeBuilder,
        Optional<Branch<Path>> branch,
        SourceFeature.NshmFault feature,
        ModelData data) {

      /* Possibly split on normal fault dip variants. */
      SourceConfig.Fault config = data.sourceConfig().asFault();
      boolean isNormal = FocalMech.fromRake(feature.rake) == FocalMech.NORMAL;

      if (config.dipTree.isPresent() && isNormal) {
        return loadNormalDipTree(treeBuilder, branch, feature, data);
      }

      return loadSingle(treeBuilder, branch, data, feature);
    }

    private SourceTree.Builder loadSingle(
        SourceTree.Builder treeBuilder,
        Optional<Branch<Path>> branch,
        ModelData data,
        SourceFeature.NshmFault feature) {

      /* If supplied branch is empty, we need to set a singleton tree root. */
      if (branch.isEmpty()) {
        LogicTree<Path> root = LogicTree.singleton(
            Deserialize.SOURCE_TREE,
            feature.name,
            Path.of(feature.name));
        treeBuilder.root(root);
        branch = Optional.of(root.get(0));
      }

      double Mw = feature.magnitude.isPresent()
          ? feature.magnitude.orElseThrow()
          : FaultRuptureSet.magnitudeWc94(feature.length);
      double Δz = feature.lowerDepth - feature.upperDepth;
      double width = computeWidth(Δz, feature.dip);
      Feature newFeature = Feature.copyOf(feature.source)
          .properties(Properties.fromFeature(feature.source)
              .put(Key.MAGNITUDE, Mw)
              .put(Key.WIDTH, width)
              .build())
          .build();
      feature = new SourceFeature.NshmFault(newFeature);

      RuptureSet<? extends Source> ruptures = createSingleRuptureSet(data, feature);
      treeBuilder.addLeaf(branch.orElseThrow(), ruptures);
      return treeBuilder;
    }

    private SourceTree.Builder loadNormalDipTree(
        SourceTree.Builder treeBuilder,
        Optional<Branch<Path>> branch, // if absent, set tree root
        SourceFeature.NshmFault feature,
        ModelData data) {

      SourceConfig.Fault config = data.sourceConfig().asFault();
      LogicTree<Double> dipTree = config.dipTree.orElseThrow();
      LogicTree<Path> dipSourceTree = ModelTrees.toPathTree(dipTree);
      DipSlipModel dipSlipModel = config.dipSlipModel.orElseThrow();

      if (branch.isPresent()) {
        treeBuilder.addBranches(branch.orElseThrow(), dipSourceTree);
      } else {
        treeBuilder.root(dipSourceTree);
      }

      /*
       * Create new features with updated properties. Magnitude may be present
       * in properties for regional, historic, or other overrides.
       */
      double Mw = feature.magnitude.isPresent()
          ? feature.magnitude.orElseThrow()
          : FaultRuptureSet.magnitudeWc94(feature.length);
      for (int i = 0; i < dipTree.size(); i++) {
        Branch<Double> dipBranch = dipTree.get(i);
        int branchDip = (int) (feature.dip + dipBranch.value());
        int rateDip = (int) ((dipSlipModel == FIXED) ? feature.dip : branchDip);
        String branchName = String.format("%s [%s°]", feature.name, branchDip);
        double Δz = feature.lowerDepth - feature.upperDepth;
        double width = computeWidth(Δz, branchDip);
        Feature dipFeature = Feature.copyOf(feature.source)
            .properties(Properties.fromFeature(feature.source)
                .put(Key.NAME, branchName)
                .put(Key.DIP, branchDip)
                .put(Key.RATE_DIP, rateDip)
                .put(Key.MAGNITUDE, Mw)
                .put(Key.WIDTH, width)
                .build())
            .build();

        RuptureSet<? extends Source> ruptures = createSingleRuptureSet(
            data,
            new SourceFeature.NshmFault(dipFeature));

        treeBuilder.addLeaf(dipSourceTree.get(i), ruptures);
      }
      return treeBuilder;// .build();
    }

    /* This exists for consistency with legacy codes */
    @Deprecated
    private static final double TO_RAD = 3.14159 / 180.0;

    @Deprecated
    private static double computeWidth(double Δz, double dip) {
      // Inexplicable rounding up of 45° where other dips go down
      RoundingMode rMode = (dip == 45.0) ? RoundingMode.CEILING : RoundingMode.FLOOR;
      return Maths.round(Δz / Math.sin(dip * TO_RAD), 5, rMode);
    }

    private FaultRuptureSet createSingleRuptureSet(
        ModelData data,
        SourceFeature.NshmFault feature) {

      Map<String, LogicTree<Mfd.Properties>> mfdMap = data.mfdMap().orElseThrow();
      // TODO singleton map is a bad assumtion,
      LogicTree<Mfd.Properties> mfdTree = mfdMap.values().iterator().next();
      data.faultFeatureMap(Map.of(feature.id, feature));

      return FaultRuptureSet.builder()
          .modelData(data)
          .name(feature.name)
          .id(feature.id)
          .sections(List.of(feature.id))
          .mfdTree(mfdTree)
          .build();
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();
      SourceTree.Info info = readTreeInfo(dir).orElseThrow(nse(dir, TREE_INFO));
      String name = info.name.orElse(checkNotNull(dir.getFileName()).toString());

      System.out.println("    tree: [" + info.id + "] " + root.relativize(dir));

      /* Set flag to process tree of single geojson sources. */
      boolean singleSourceTree = tree.get(0).value().toString().endsWith(".geojson");

      /* Set flag to process a logic-tree of fault-system solutions. */
      var faultSections = readFaultFeatures(dir);
      faultSections.ifPresent(data::faultFeatureMap);
      boolean faultSystemTree = faultSections.isEmpty() && !singleSourceTree;

      /*
       * Set data flag that we're processing a source tree to correctly build
       * faultRuptureSet MFDs downstream. We want a tree of singleSources to
       * adhere to a fault rate-tree, if present, but we want it ignored if
       * we're in a fully fledged source tree (e.g. Wasatch).
       */
      if (!singleSourceTree && !faultSystemTree) {
        data.faultSourceTree();
      }

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(name)
          .id(info.id)
          .setting(data.tectonicSetting())
          .type(faultSystemTree ? FAULT_SYSTEM : FAULT)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        if (singleSourceTree) {
          processSingleBranch(branch, treeBuilder, data);
        } else if (faultSystemTree) {
          processSystemBranch(branch, treeBuilder, data);
        } else {
          processBranch(branch, treeBuilder, data);
        }
      }
      return treeBuilder.build();
    }

    /* Process fault source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch:        " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.err.println("                 TODO: null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      readMfdConfig(dir).ifPresent(data::mfdConfig);

      /*
       * Don't use loadRateTree() here; nested rate trees on branches are more
       * likely numeric recurrence rates than slip or cluster rate branches.
       */
      readRateTree(dir).ifPresent(data::rateTree);

      /* Can be source-tree or source-group. */
      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {

        /* Could have fault or cluster rupture-set. */
        var frs = readFaultRuptureSet(dir, data);
        if (frs.isPresent()) {
          treeBuilder.addLeaf(branch, frs.orElseThrow());
        } else {
          var crsd = readFaultClusterSet(dir, data).orElseThrow();
          processClusterBranch(branch, treeBuilder, data, crsd);
        }
      }
    }

    /*
     * Process fault clusters adding additional cluster rate branches. The
     * supplied cluster-set is used as a model to create a tree of rate
     * branches, each with the same cluster geometry.
     */
    private void processClusterBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData data,
        ClusterRuptureSet.Data clusterData) {

      LogicTree<Double> rateTree = data.rateTree().orElseThrow();
      LogicTree.Builder<Path> pathTreeBuilder = LogicTree.builder(rateTree.name());
      List<ClusterRuptureSet> crsList = new ArrayList<>();

      for (int i = 0; i < rateTree.size(); i++) {

        Branch<Double> rateBranch = rateTree.get(i);
        double rate = (rateBranch.value() <= 0.0) ? 0.0 : rateBranch.value();
        String rateLabel = rateBranch.id() + ((rate == 0.0)
            ? "-null"
            : "-" + String.valueOf((int) Math.round(1.0 / rate)) + "-yr");
        String clusterLabel = clusterData.name + " : " + rateLabel;

        Path branchPath = root.relativize(branch.value()).resolve(rateLabel);
        System.out.println("  branch:        " + branchPath);

        ClusterSource clusterSource = new ClusterSource.Builder()
            .name(clusterLabel)
            .id(clusterData.id)
            .rate(rateBranch.value()) // cluster rate
            .faults(clusterData.faultRuptureSets)
            .buildClusterSource();

        ClusterRuptureSet crs = ClusterRuptureSet.builder()
            .name(clusterLabel)
            .id(clusterData.id)
            .setSource(clusterSource)
            .modelData(data)
            .build();
        crsList.add(crs);

        pathTreeBuilder.addBranch(
            clusterLabel,
            Path.of(clusterLabel),
            rateBranch.weight());
      }

      LogicTree<Path> pathTree = pathTreeBuilder.build();
      treeBuilder.addBranches(branch, pathTree);

      for (int i = 0; i < pathTree.size(); i++) {
        treeBuilder.addLeaf(pathTree.get(i), crsList.get(i));
      }
    }

    /* Process fault system source tree branches. */
    private void processSystemBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData data) {

      /*
       * No configuration overrides are currently expected for
       * fault-system-solutions so this is simpler than above.
       */
      Path dir = branch.value();
      System.out.println("  branch:        " + root.relativize(dir));
      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {
        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);
        for (Branch<Path> child : children) {
          processSystemBranch(child, treeBuilder, data);
        }
      } else {
        treeBuilder.addLeaf(
            branch,
            readSystemRuptureSet(dir, data).orElseThrow());
      }
    }

    /*
     * Process branches consisting of individual geojson files as would be
     * processed using loadSingleSource().
     */
    private void processSingleBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData data) {

      if (checkNotNull(branch.value().getFileName()).toString().equals("null")) {
        System.err.println("                 TODO: null branch");
        return;
      }

      SourceFeature.NshmFault feature = SourceFeature.newNshmFault(branch.value());
      loadSingleSource(treeBuilder, Optional.of(branch), feature, data);
    }
  }

  static List<SourceTree> decollementSources(
      Path dir,
      TectonicSetting setting,
      Optional<GmmSet> gmms) {

    GmmSet sourceGmms = readGmms(dir).or(() -> gmms).orElseThrow();
    SourceConfig srcConfig = readDecollementConfig(dir).orElseThrow();
    ModelData data = ModelData.create(setting, sourceGmms, srcConfig);
    return new Decollement(dir).loadSourceDirectory(dir, data);
  }

  private static class Decollement extends ModelLoader {

    Decollement(Path root) {
      super(root);
      logRoot(root, DECOLLEMENT);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readDecollementConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readRateTree(dir).ifPresent(data::rateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      throw new UnsupportedOperationException(
          "Wrap single décollement sources in a single branch logic tree");
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();
      SourceTree.Info info = readTreeInfo(dir).orElseThrow(nse(dir, TREE_INFO));
      String name = info.name.orElse(checkNotNull(dir.getFileName()).toString());

      System.out.println("    tree: [" + info.id + "] " + root.relativize(dir));

      data.interfaceFeatureMap(readInterfaceFeatures(dir).orElseThrow());

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(name)
          .id(info.id)
          .setting(data.tectonicSetting())
          .type(DECOLLEMENT)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        processBranch(branch, treeBuilder, ModelData.copyOf(data));
      }

      return treeBuilder.build();
    }

    /* Process interface source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch:        " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("                 TODO null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      readMfdConfig(dir).ifPresent(data::mfdConfig);
      readRateTree(dir).ifPresent(data::rateTree);

      /* Can be source-tree or source-group. */
      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {
        treeBuilder.addLeaf(
            branch,
            readInterfaceRuptureSet(dir, data).orElseThrow());
      }
    }
  }

  static List<SourceTree> interfaceSources(
      Path dir,
      TectonicSetting setting,
      Optional<GmmSet> gmms) {

    GmmSet sourceGmms = readGmms(dir).or(() -> gmms).orElseThrow();
    SourceConfig srcConfig = readInterfaceConfig(dir).orElseThrow();
    ModelData data = ModelData.create(setting, sourceGmms, srcConfig);
    return new Interface(dir).loadSourceDirectory(dir, data);
  }

  private static class Interface extends ModelLoader {

    Interface(Path root) {
      super(root);
      logRoot(root, INTERFACE);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readInterfaceConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readRateTree(dir).ifPresent(data::rateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      throw new UnsupportedOperationException(
          "Wrap single interface sources in a single branch logic tree");
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();
      SourceTree.Info info = readTreeInfo(dir).orElseThrow(nse(dir, TREE_INFO));
      String name = info.name.orElse(checkNotNull(dir.getFileName()).toString());

      System.out.println("    tree: [" + info.id + "] " + root.relativize(dir));

      data.interfaceFeatureMap(readInterfaceFeatures(dir).orElseThrow());

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(name)
          .id(info.id)
          .setting(data.tectonicSetting())
          .type(INTERFACE)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        processBranch(branch, treeBuilder, ModelData.copyOf(data));
      }

      return treeBuilder.build();
    }

    /* Process interface source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch:        " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("                 TODO: null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      readRateTree(dir).ifPresent(data::rateTree);

      /* Can be source-tree or source-group. */
      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {
        treeBuilder.addLeaf(
            branch,
            readInterfaceRuptureSet(dir, data).orElseThrow());
      }
    }
  }

  static List<SourceTree> gridSources(
      Path dir,
      TectonicSetting setting,
      Optional<GmmSet> gmms) {

    GmmSet sourceGmms = readGmms(dir).or(() -> gmms).orElseThrow();
    SourceConfig srcConfig = readGridConfig(dir).orElseThrow();
    ModelData data = ModelData.create(setting, sourceGmms, srcConfig);
    /* Grid data directory required at grid-sources root. */
    Path gridDataDir = checkGridDataDirectory(dir).orElseThrow();
    data.gridDataDirectory(gridDataDir);
    /* Grid features directory required at grid-sources root. */
    var gridFeatureMap = readGridFeatures(dir).orElseThrow();
    data.gridFeatureMap(gridFeatureMap);
    return new Grid(dir).loadSourceDirectory(dir, data);
  }

  private static class Grid extends ModelLoader {

    Grid(Path root) {
      super(root);
      logRoot(root, GRID);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readGridConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readRateTree(dir).ifPresent(data::rateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      throw new UnsupportedOperationException(
          "Wrap single grids in a single branch logic tree");
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();
      SourceTree.Info info = readTreeInfo(dir).orElseThrow(nse(dir, TREE_INFO));
      String name = info.name.orElse(checkNotNull(dir.getFileName()).toString());

      System.out.println("    tree: [" + info.id + "] " + name);

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(name)
          .id(info.id)
          .setting(data.tectonicSetting())
          .type(GRID)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        processBranch(branch, treeBuilder, data);
      }

      return treeBuilder.build();
    }

    /* Process grid source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch:        " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("                 TODO: null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      loadRateTree(dir, data);
      loadSourceConfig(dir, data);

      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {

        /* Expect a rupture-sets file. */

        List<? extends RuptureSet<? extends Source>> ruptureSets =
            readGridRuptureSets(dir, data).orElseThrow();

        /* Attach singleton as leaf. */
        if (ruptureSets.size() == 1) {
          treeBuilder.addLeaf(branch, ruptureSets.get(0));
          return;
        }

        /* Otherwise create logic group of rupture set pseudo paths. */
        LogicGroup.Builder<Path> grsGroup = LogicGroup.builder();
        for (RuptureSet<? extends Source> ruptureSet : ruptureSets) {
          grsGroup.addBranch(ruptureSet.name(), Path.of(ruptureSet.name()), 1.0);
        }
        LogicTree<Path> grsTree = grsGroup.build();
        treeBuilder.addBranches(branch, grsTree);

        for (int j = 0; j < grsTree.size(); j++) {
          treeBuilder.addLeaf(grsTree.get(j), ruptureSets.get(j));
        }
      }
    }
  }

  static List<SourceTree> slabSources(
      Path dir,
      TectonicSetting setting,
      Optional<GmmSet> gmms) {

    GmmSet sourceGmms = readGmms(dir).or(() -> gmms).orElseThrow();
    SourceConfig srcConfig = readSlabConfig(dir).orElseThrow();
    ModelData data = ModelData.create(setting, sourceGmms, srcConfig);
    /* Grid data directory required at subduction-slab root. */
    Path gridDataDir = checkGridDataDirectory(dir).orElseThrow();
    data.gridDataDirectory(gridDataDir);
    /* Grid features directory required at subduction-slab root. */
    var gridFeatureMap = readGridFeatures(dir).orElseThrow();
    data.gridFeatureMap(gridFeatureMap);
    return new Slab(dir).loadSourceDirectory(dir, data);
  }

  private static class Slab extends ModelLoader {

    Slab(Path root) {
      super(root);
      logRoot(root, SLAB);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readSlabConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readRateTree(dir).ifPresent(data::rateTree);
      // readGridRatesTree(dir).ifPresent(data::gridRatesTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      throw new UnsupportedOperationException(
          "Wrap single grids in a single branch logic tree");
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();
      SourceTree.Info info = readTreeInfo(dir).orElseThrow(nse(dir, TREE_INFO));
      String name = info.name.orElse(checkNotNull(dir.getFileName()).toString());

      System.out.println("    tree: [" + info.id + "] " + root.relativize(dir));

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(name)
          .id(info.id)
          .setting(data.tectonicSetting())
          .type(SLAB)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        processBranch(branch, treeBuilder, data);
      }

      return treeBuilder.build();
    }

    /* Process intraslab source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch:        " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("                 TODO: null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      loadRateTree(dir, data);
      loadSourceConfig(dir, data);

      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {

        /* Expect a rupture-sets file. */

        List<? extends RuptureSet<? extends Source>> ruptureSets =
            readSlabRuptureSets(dir, data).orElseThrow();

        /* Attach singleton as leaf. */
        if (ruptureSets.size() == 1) {
          treeBuilder.addLeaf(branch, ruptureSets.get(0));
          return;
        }

        /* Otherwise create logic group of rupture set pseudo paths. */
        LogicGroup.Builder<Path> grsGroup = LogicGroup.builder();
        for (RuptureSet<? extends Source> ruptureSet : ruptureSets) {
          grsGroup.addBranch(ruptureSet.name(), Path.of(ruptureSet.name()), 1.0);
        }
        LogicTree<Path> grsTree = grsGroup.build();
        treeBuilder.addBranches(branch, grsTree);

        for (int j = 0; j < grsTree.size(); j++) {
          treeBuilder.addLeaf(grsTree.get(j), ruptureSets.get(j));
        }
      }
    }
  }

  static List<SourceTree> zoneSources(
      Path dir,
      TectonicSetting setting,
      Optional<GmmSet> gmms) {

    GmmSet sourceGmms = readGmms(dir).or(() -> gmms).orElseThrow();
    SourceConfig srcConfig = readZoneConfig(dir).orElseThrow();
    ModelData data = ModelData.create(setting, sourceGmms, srcConfig);
    return new Zone(dir).loadSourceDirectory(dir, data);
  }

  private static class Zone extends ModelLoader {

    Zone(Path root) {
      super(root);
      logRoot(root, ZONE);
    }

    @Override
    void loadSourceConfig(Path dir, ModelData data) {
      readZoneConfig(dir).ifPresent(data::sourceConfig);
    }

    @Override
    void loadRateTree(Path dir, ModelData data) {
      readRateTree(dir).ifPresent(data::rateTree);
    }

    @Override
    SourceTree loadSingleSource(
        Path geojson,
        ModelData data) {

      SourceFeature.Zone feature = SourceFeature.newSourceZone(geojson);
      System.out.println("    tree: [" + feature.id + "] " + root.relativize(geojson));

      Properties props = feature.source.properties();
      String ratesFile = props.getString("rate").orElseThrow();

      ZoneRuptureSet zrs = ZoneRuptureSet.builder()
          .feature(feature)
          .modelData(data)
          .rates(geojson.resolveSibling(ratesFile))
          .mfdTree(Deserialize.mfdTree(props, data))
          .build();

      LogicTree<Path> root = LogicTree.singleton(
          Deserialize.SOURCE_TREE,
          feature.name,
          Path.of(feature.name));

      SourceTree tree = SourceTree.builder()
          .name(feature.name)
          .id(feature.id)
          .setting(data.tectonicSetting())
          .type(ZONE)
          .gmms(data.gmms())
          .root(root)
          .addLeaf(root.get(0), zrs)
          .build();

      return tree;
    }

    @Override
    SourceTree loadSourceTree(
        Path dir,
        ModelData data) {

      LogicTree<Path> tree = readSourceTree(dir).orElseThrow();
      SourceTree.Info info = readTreeInfo(dir).orElseThrow(nse(dir, TREE_INFO));
      String name = info.name.orElse(checkNotNull(dir.getFileName()).toString());

      System.out.println("    tree: [" + info.id + "] " + root.relativize(dir));

      SourceTree.Builder treeBuilder = SourceTree.builder()
          .name(name)
          .id(info.id)
          .setting(data.tectonicSetting())
          .type(ZONE)
          .gmms(data.gmms())
          .root(tree);

      for (Branch<Path> branch : tree) {
        processBranch(branch, treeBuilder, data);
      }

      return treeBuilder.build();
    }

    /* Process zone (area) source tree branches. */
    private void processBranch(
        Branch<Path> branch,
        SourceTree.Builder treeBuilder,
        ModelData dataIn) {

      Path dir = branch.value();
      System.out.println("  branch:        " + root.relativize(dir));

      if (checkNotNull(dir.getFileName()).toString().equals("null")) {
        System.out.println("                 TODO: null branch");
        return;
      }

      ModelData data = ModelData.copyOf(dataIn);
      readRateTree(dir).ifPresent(data::rateTree);

      Optional<LogicTree<Path>> tree = readSourceTree(dir);
      if (tree.isPresent()) {

        LogicTree<Path> children = tree.orElseThrow();
        treeBuilder.addBranches(branch, children);

        for (Branch<Path> child : children) {
          processBranch(child, treeBuilder, data);
        }

      } else {

        Path zonePath;
        try (DirectoryStream<Path> sourceStream = newDirectoryStream(branch.value(), "*.geojson")) {
          zonePath = Iterables.getOnlyElement(sourceStream);
        } catch (IOException ioe) {
          throw new RuntimeException(ioe);
        }

        SourceFeature.Zone zone = SourceFeature.newSourceZone(zonePath);

        Properties props = zone.source.properties();
        String ratesFile = props.getString("rate").orElseThrow();

        ZoneRuptureSet zrs = ZoneRuptureSet.builder()
            .feature(zone)
            .modelData(data)
            .rates(zonePath.resolveSibling(ratesFile))
            .mfdTree(Deserialize.mfdTree(props, data))
            .build();

        treeBuilder.addLeaf(branch, zrs);
      }
    }
  }

}
