package gov.usgs.earthquake.nshmp.model;

import static gov.usgs.earthquake.nshmp.Text.NEWLINE;
import static gov.usgs.earthquake.nshmp.model.Deserialize.ID;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MODEL;
import static gov.usgs.earthquake.nshmp.model.Deserialize.NAME;
import static gov.usgs.earthquake.nshmp.model.Deserialize.SPACING;
import static java.util.stream.Collectors.toUnmodifiableMap;
import static java.util.stream.Collectors.toUnmodifiableSet;

import java.awt.geom.Area;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.OptionalDouble;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.Properties;

/**
 * Site data wrapper class. This class provides one method {@code get(Location)}
 * that returns the superset of all site data parameters required by the current
 * suite of supported NSHMs.
 *
 * This class assumes that if a location 'hit test' succeeds on an area for
 * which there are site data, then non null values will exist for the location.
 *
 * @author U.S. Geological Survey
 */
public class SiteData {

  /** Empty site data instance. */
  public static final SiteData EMPTY = new SiteData(null, null) {
    @Override
    public Values get(Location loc) {
      return Values.EMPTY;
    }
  };

  private static Path BASIN_DIR = Path.of("basin");
  private static Path MARGIN_DIR = Path.of("margin");

  private Set<Basin> basins;
  private Set<Margin> margins;

  private SiteData(Set<Basin> basins, Set<Margin> margins) {
    this.basins = basins;
    this.margins = margins;
  }

  static SiteData create(Path dir) {
    Set<Basin> basins = init(dir.resolve(BASIN_DIR), Basin::new);
    Set<Margin> margins = init(dir.resolve(MARGIN_DIR), Margin::new);
    return (basins.isEmpty() && margins.isEmpty())
        ? EMPTY
        : new SiteData(basins, margins);
  }

  /**
   * Return the site data values for the specified location. Once a polygon
   * 'hit' ocurs, method breaks out of search loops.
   *
   * @param location of interest
   */
  public Values get(Location location) {
    Values.Builder builder = new Values.Builder();
    for (Basin basin : basins) {
      Location snapped = snapToGrid(location, basin.spacing, basin.scale);
      if (basin.contains(snapped)) {
        builder.basinValues(basin.data.get(snapped));
        break;
      }
    }
    for (Margin margin : margins) {
      Location snapped = snapToGrid(location, margin.spacing, margin.scale);
      if (margin.contains(snapped)) {
        builder.marginValues(margin.data.get(snapped));
        break;
      }
    }
    return builder.build();
  }

  /* Snap location to closest multiple of spacing. */
  private static Location snapToGrid(Location loc, double spacing, int scale) {
    return Location.create(
        snapToGrid(loc.longitude, spacing, scale),
        snapToGrid(loc.latitude, spacing, scale));
  }

  /*
   * Snap value to closest multiple of spacing.
   *
   * Double precision rounding errors results in grid midpoints rounding both up
   * and down in the initial Math.round() call.
   */
  private static double snapToGrid(double value, double spacing, int scale) {
    return Maths.round(
        Math.round(value / spacing) * spacing,
        scale, RoundingMode.HALF_UP);
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder(ModelFiles.SITE_DATA_DIR).append("/");
    if (this == EMPTY) {
      return sb.append(" (empty)").toString();
    }
    sb.append(NEWLINE);
    basins.forEach(basin -> sb.append("   basin: ").append(basin.name).append(NEWLINE));
    margins.forEach(margin -> sb.append("  margin: ").append(margin.name).append(NEWLINE));
    return sb.toString();
  }

  /** Optional site data values associated with a location. */
  public static class Values {

    private static final Values EMPTY = new Builder().build();

    /** Depth to the shear-wave velocity horizon of 1.0 km/s, in km. */
    public final OptionalDouble z1p0;

    /** Depth to the shear-wave velocity horizon of 2.5 km/s, in km. */
    public final OptionalDouble z2p5;

    /** Cpntinental margin sediment thickness, in km. */
    public final OptionalDouble zSed;

    private Values(Builder builder) {
      this.z1p0 = builder.z1p0;
      this.z2p5 = builder.z2p5;
      this.zSed = builder.zSed;
    }

    private static class Builder {

      OptionalDouble z1p0 = OptionalDouble.empty();
      OptionalDouble z2p5 = OptionalDouble.empty();
      OptionalDouble zSed = OptionalDouble.empty();

      Builder basinValues(Basin.Values values) {
        this.z1p0 = OptionalDouble.of(values.z1p0);
        this.z2p5 = OptionalDouble.of(values.z2p5);
        return this;
      }

      Builder marginValues(Margin.Values values) {
        this.zSed = OptionalDouble.of(values.zSed);
        return this;
      }

      Values build() {
        return new Values(this);
      }
    }
  }

  private static Location toLocation(Record record) {
    return Location.create(
        record.getDouble("lon"),
        record.getDouble("lat"));
  }

  private static <T> Set<T> init(Path dir, Function<Path, T> mapper) {
    if (Files.exists(dir)) {
      try (DirectoryStream<Path> ds = Files.newDirectoryStream(dir, "*.geojson")) {
        return StreamSupport.stream(ds.spliterator(), false)
            .map(mapper)
            .collect(toUnmodifiableSet());
      } catch (IOException ioe) {
        throw new RuntimeException(ioe);
      }
    }
    return Set.of();
  }

  /* Container for a basin region and depth data. */
  static final class Basin {

    final String name;
    final String id;
    final String model;
    final double spacing;
    final int scale;
    final Feature feature;
    final Area area;
    final Map<Location, Values> data;

    private Basin(Path geojson) {
      Feature feature = GeoJson.from(geojson).toFeature();
      Properties properties = feature.properties();
      this.name = properties.getString(NAME).orElseThrow();
      this.id = properties.getString(ID).orElseThrow();
      this.model = properties.getString(MODEL).orElseThrow();
      this.spacing = properties.getDouble(SPACING).orElseThrow();
      this.scale = BigDecimal.valueOf(spacing).scale();
      this.feature = feature;
      this.area = new Area(Locations.toPath(feature.asPolygonBorder()));
      Path csvpath = geojson.resolveSibling(id + ".csv");
      this.data = loadData(csvpath);
    }

    /* Return whether the basin contains the supplied location. */
    boolean contains(Location loc) {
      return area.contains(loc.longitude, loc.latitude);
    }

    private Map<Location, Values> loadData(Path file) {
      DelimitedData csv = DelimitedData.comma(file);
      try (Stream<Record> stream = csv.records()) {
        return stream.collect(toUnmodifiableMap(
            SiteData::toLocation,
            Basin::toValues));
      }
    }

    private static Values toValues(Record record) {
      return new Values(
          record.getDouble("z1p0"),
          record.getDouble("z2p5"));
    }

    static final class Values {

      final double z1p0;
      final double z2p5;

      Values(double z1p0, double z2p5) {
        this.z1p0 = z1p0;
        this.z2p5 = z2p5;
      }
    }
  }

  /* Container for continental margin sediment thickness data. */
  static final class Margin {

    final String name;
    final String id;
    final String model;
    final double spacing;
    final int scale;
    final Feature feature;
    final Area area;
    final Map<Location, Values> data;

    private Margin(Path geojson) {
      Feature feature = GeoJson.from(geojson).toFeature();
      Properties properties = feature.properties();
      this.name = properties.getString(NAME).orElseThrow();
      this.id = properties.getString(ID).orElseThrow();
      this.model = properties.getString(MODEL).orElseThrow();
      this.spacing = properties.getDouble(SPACING).orElseThrow();
      this.scale = BigDecimal.valueOf(spacing).scale();
      this.feature = feature;
      this.area = new Area(Locations.toPath(feature.asPolygonBorder()));
      Path csvpath = geojson.resolveSibling(id + ".csv");
      this.data = loadData(csvpath);
    }

    /* Return whether the basin contains the supplied location. */
    boolean contains(Location loc) {
      return area.contains(loc.longitude, loc.latitude);
    }

    private Map<Location, Values> loadData(Path file) {
      DelimitedData csv = DelimitedData.comma(file);
      try (Stream<Record> stream = csv.records()) {
        return stream.collect(toUnmodifiableMap(
            SiteData::toLocation,
            Margin::toValues));
      }
    }

    private static Values toValues(Record record) {
      double value = Maths.round(record.getDouble("zsed") / 1000.0, 3);
      return new Values(value);
    }

    static final class Values {

      final double zSed;

      Values(double zSed) {
        this.zSed = zSed;
      }
    }
  }

}
